# A Python program for Prims's MST for  
# adjacency list representation of graph 
  
import math
import numpy as np
from collections import defaultdict


# Converts from adjacency matrix to adjacency list 
def convert_matrix_to_list(matrix):
    adjList = defaultdict(list)
    for i in range(len(matrix)):
        for j in range(len(matrix[i])):
            if matrix[i][j] > 0:
                newNode = [j, matrix[i][j]] 
                adjList[i].append(newNode)
    return adjList


class Heap(): 
    def __init__(self): 
        self.array = [] 
        self.size = 0
        self.pos = [] 
  
    def newMinHeapNode(self, v, dist): 
        minHeapNode = [v, dist] 
        return minHeapNode 
  
    # A utility function to swap two nodes of  
    # min heap. Needed for min heapify 
    def swapMinHeapNode(self, a, b): 
        t = self.array[a] 
        self.array[a] = self.array[b] 
        self.array[b] = t 
  
    # A standard function to heapify at given idx 
    # This function also updates position of nodes  
    # when they are swapped. Position is needed  
    # for decreaseKey() 
    def minHeapify(self, idx): 
        smallest = idx 
        left = 2 * idx + 1
        right = 2 * idx + 2
  
        if left < self.size and self.array[left][1] < self.array[smallest][1]: 
            smallest = left 
  
        if right < self.size and self.array[right][1] < self.array[smallest][1]: 
            smallest = right 
  
        # The nodes to be swapped in min heap  
        # if idx is not smallest 
        if smallest != idx: 
            # Swap positions 
            self.pos[ self.array[smallest][0] ] = idx 
            self.pos[ self.array[idx][0] ] = smallest 
  
            # Swap nodes 
            self.swapMinHeapNode(smallest, idx) 
  
            self.minHeapify(smallest) 
  
    # Standard function to extract minimum node from heap 
    def extractMin(self): 
        # Return NULL wif heap is empty 
        if self.isEmpty() == True: 
            return
  
        # Store the root node 
        root = self.array[0] 
  
        # Replace root node with last node 
        lastNode = self.array[self.size - 1] 
        self.array[0] = lastNode 
  
        # Update position of last node 
        self.pos[lastNode[0]] = 0
        self.pos[root[0]] = self.size - 1
  
        # Reduce heap size and heapify root 
        self.size -= 1
        self.minHeapify(0) 
  
        return root 
  
    def isEmpty(self): 
        return True if self.size == 0 else False
  
    def decreaseKey(self, v, dist): 
        # Get the index of v in  heap array
        i = self.pos[v] 
  
        # Get the node and update its dist value 
        self.array[i][1] = dist 
  
        # Travel up while the complete tree is not  
        # hepified. This is a O(Logn) loop 
        while i > 0 and self.array[i][1] < self.array[(i - 1) // 2][1]:
            # Swap this node with its parent 
            self.pos[ self.array[i][0] ] = (i-1) // 2
            self.pos[ self.array[(i-1)//2][0]] = i 
            self.swapMinHeapNode(i, (i - 1) // 2) 
  
            # move to parent index 
            i = (i - 1) // 2; 

    # A utility function to check if a given vertex 
    # 'v' is in min heap or not 
    def isInMinHeap(self, v): 
        if self.pos[v] < self.size: 
            return True
        return False


class Graph(): 
    def __init__(self, adjacency_matrix): 
        self.V = adjacency_matrix.shape[0] 
        self._graph = convert_matrix_to_list(adjacency_matrix)

    def _build_edges(self, parent, n):
        edges = []
        for i in range(1, n): 
            edges.append([parent[i], i])
        return np.array(edges)

    # The main function that prints the Minimum  
    # Spanning Tree(MST) using the Prim's Algorithm.  
    # It is a O(ELogV) function 
    def PrimMST(self): 
        # Get the number of vertices in graph 
        V = self.V   
          
        # key values used to pick minimum weight edge in cut 
        key = []    
          
        # List to store contructed MST 
        parent = []  
  
        # minHeap represents set E 
        minHeap = Heap() 
  
        # Initialize min heap with all vertices. Key values of all 
        # vertices (except the 0th vertex) is is initially infinite 
        for v in range(V): 
            parent.append(-1) 
            key.append(math.inf) 
            minHeap.array.append(minHeap.newMinHeapNode(v, key[v])) 
            minHeap.pos.append(v) 
  
        # Make key value of 0th vertex as 0 so  
        # that it is extracted first 
        minHeap.pos[0] = 0
        key[0] = 0
        minHeap.decreaseKey(0, key[0]) 
  
        # Initially size of min heap is equal to V 
        minHeap.size = V
  
        # In the following loop, min heap contains all nodes 
        # not yet added in the MST. 
        while minHeap.isEmpty() == False: 
            # Extract the vertex with minimum distance value 
            newHeapNode = minHeap.extractMin() 
            u = newHeapNode[0] 
  
            # Traverse through all adjacent vertices of u  
            # (the extracted vertex) and update their  
            # distance values 
            for pCrawl in self._graph[u]:
                v = pCrawl[0] 
  
                # If shortest distance to v is not finalized  
                # yet, and distance to v through u is less than 
                # its previously calculated distance 
                if minHeap.isInMinHeap(v) and pCrawl[1] < key[v]: 
                    key[v] = pCrawl[1] 
                    parent[v] = u 
  
                    # update distance value in min heap also 
                    minHeap.decreaseKey(v, key[v]) 
  
        return self._build_edges(parent, V)


def MST(adjacency_matrix):
    graph = Graph(adjacency_matrix)
    return graph.PrimMST()

# Debug
# if __name__ == "__main__":
#     a = np.array(
#           [[0, 9, 75, 2, 0, 10],
#           [9, 0, 95, 19, 42, 99],
#           [75, 95, 0, 51, 66, 12],
#           [2, 19, 51, 0, 31, 22],
#           [0, 42, 66, 31, 0, 89],
#           [10, 99, 12, 22, 89, 0]]
#         )

#     graph = Graph(a)
#     print("Adjacency List:") 
#     for i in graph._graph: 
#         print(i, end ="") 
#         for node in graph._graph[i]: 
#             print(" -> {}".format(node[0]), end ="") 
#         print()

#     print("="*20)
#     MST = graph.PrimMST()
#     print(MST)



import math
import numpy as np
from collections import defaultdict
 
# Converts from adjacency matrix to adjacency list 
def convert_matrix_to_list(a):
    adjList = defaultdict(list)
    for i in range(len(a)):
        for j in range(len(a[i])):
            if a[i][j] > 0:
                adjList[i].append(j)
    return adjList


def MST(G):
    """
    Find a Minimum Spanning Tree using Prim's Algorithm
        - G: an adjacency matrix representing a graph
    """

    tree = []

    INF = math.inf

    # Number of vertices in graph
    V = G.shape[-1]

    # Create a set S to track selected vertices
    # selected vertices will become true otherwise false
    S = np.zeros(V, dtype=bool)

    # Initialize number of edge to 0
    n_edges = 0

    # The number of edges in Minimum Spanning Tree will be
    # always less than (V - 1), where V is the number of vertices in the graph

    # Choose 0th vertex and make it true
    S[0] = True

    # For every selected vertex in the set S, find all adjacent vertices 
    # that are not included in the set S.
    # Choose an adjacent vertex that is nearest to one of selected vertices.
    while (n_edges < V - 1):
        minimum = INF
        x = 0
        y = 0
        for i in range(V):
            if S[i]:
                for j in range(V):
                    if ((not S[j]) and G[i][j]):
                        if minimum > G[i][j]:
                            minimum = G[i][j]
                            x = i
                            y = j
        S[y] = True
        tree.append([x, y])
        n_edges += 1
    return np.array(tree)


if __name__ == "__main__":
    a = np.array(
          [[0, 9, 75, 2, 0, 10],
          [9, 0, 95, 19, 42, 99],
          [75, 95, 0, 51, 66, 12],
          [2, 19, 51, 0, 31, 22],
          [0, 42, 66, 31, 0, 89],
          [10, 99, 12, 22, 89, 0]]
        )
    adjList = convert_matrix_to_list(a)
    print("Adjacency List:") 
    for i in adjList: 
        print(i, end ="") 
        for j in adjList[i]: 
            print(" -> {}".format(j), end ="") 
        print()

    print(MST(a))



import os
import math
import gower
import logging
import numpy as np
import collections

from dcm.logger import logger
from dcm.fast_mst import MST

DEBUG = os.environ.get("DCM_DEBUG", 0)

def F1(X, y):
    """
    Calculate Maximum Fisher's Discriminant Ratio (F1)
      - X: ndarray features
      - y: ndarray target
    """
    maxr = -math.inf
    d_ratios = {}
    X_classes = {}
    averages_classes = {}
    averages = np.average(X, axis=0)
    counter = collections.Counter(y)
    for c in counter.keys():
        X_classes[c] = X[y==c]
        averages_classes[c] = np.average(X_classes[c], axis=0)
    for i in range(X.shape[-1]):
        numerator = 0.0
        denominator = 0.0
        for c in counter.keys():
            X_c = X_classes[c]
            averages_c = averages_classes[c]
            m = (averages_c[i] - averages[i])**2
            numerator += counter[c] * m
            x = X_c[:, i]
            x = x - averages_c[i]
            x = x**2
            denominator += np.sum(x)
        r = (numerator / denominator) if denominator > 0 else 0
        d_ratios[i] = r
        maxr = r if r > maxr else maxr
        if DEBUG:
            logger.debug("{} => numerator = {}, denominator = {}, r = {}".format(i, numerator, denominator, r))
    return d_ratios, 1 / (1 + maxr)


def N1(X, y, cat_features=[]):
    """
    Calculate Fraction of Borderline Points (N1)
      - X: ndarray features
      - y: ndarray target
      - cat_features: a boolean array that specifies categorical features
    """

    if len(cat_features) == 0:
        cat_features = np.zeros(X.shape[-1], dtype=bool)
    
    # Calculate Gower distance matrix
    distance_matrix = gower.gower_matrix(X, cat_features=cat_features)

    # Generate a Minimum Spanning Tree
    tree = MST(distance_matrix)
    sub = tree[y[tree[:, 0]] != y[tree[:, 1]]]
    vertices = np.unique(sub.flatten())
    return len(vertices) / X.shape[0]


def C12(X, y):
    """
    Calculate Entropy of Class Proportions (C1) and Imbalance Ratio (C2)
      - X: ndarray features
      - y: ndarray target
    """
    n = len(y)
    counter = collections.Counter(y)
    entroy = 0.0
    tmp_sum = 0.0
    for c in counter.keys():
        pc = counter[c] / n
        entroy += pc * math.log(pc)
        tmp_sum += counter[c] / (n - counter[c])
    C1 = 1 - (-entroy / math.log(len(counter)))
    IR = ((len(counter) - 1) / len(counter)) * tmp_sum
    C2 = 1 - (1 / IR)
    return C1, C2



import os
import requests
from tqdm import tqdm

def download_creditcard(dataset_file):
  if os.path.isfile(dataset_file) and os.stat(dataset_file).st_size > 0:
    return
  url = "https://storage.googleapis.com/download.tensorflow.org/data/creditcard.csv"
  response = requests.get(url, stream=True)
  response.raise_for_status()
  filesize = int(response.headers.get('content-length', 0))
  p = tqdm(total=filesize, unit='iB', unit_scale=True)
  with open(dataset_file, "wb") as fd:
    for chunk in tqdm(response.iter_content(chunk_size=1024)):
      if chunk:
        fd.write(chunk)
        p.update(len(chunk))
  p.close()
  if filesize != 0 and p.n != filesize:
    print("ERROR, something went wrong")


import json
import pandas as pd
from dcm import dcm
import datasets

dataset_file = 'examples/datasets/creditcard.csv'
datasets.download_creditcard(dataset_file)

print("Loading dataset ... ", end="", flush=True)
dataset = pd.read_csv(dataset_file)
X = dataset.drop(columns=['Class']).values
y = dataset['Class'].values
print("DONE")

print("Calculating F1 ... ", end="", flush=True)
ratios, F1 = dcm.F1(X, y)
print("DONE")
print("F1 = {}".format(F1))
print("Discriminant Ratios = {}".format(json.dumps(ratios, indent=2)))



import logging

MSG_FORMAT = '%(asctime)s %(levelname)s %(name)s: %(message)s'
DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'
logging.basicConfig(format=MSG_FORMAT, datefmt=DATETIME_FORMAT, level=logging.DEBUG)
logger = logging.getLogger('dcm')



import setuptools
import dcm

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="data-complexity",
    version=dcm.VERSION,
    author="Son Nguyen",
    license="MIT",
    author_email="ngocson2vn@gmail.com",
    description="Data Complexity Measures",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ngocson2vn/data-complexity",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "numpy",
        "gower"
    ]
)
